#include "llvm/IR/Constants.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/Module.h"
#include "llvm/IR/Type.h"

using namespace llvm;

namespace cot {

struct HelloLLVM : public ModulePass {
public:
	static char ID;

public:
	HelloLLVM() : ModulePass(ID) { }

public:
	// This member function must implement the code of your pass.
	virtual bool runOnModule(Module &Mod) override;

	// The getAnalysisUsage allows to tell LLVM pass manager which analysis are
	// used by the pass. It is also used to declare which analysis are preserved
	// by the pass.
	virtual void getAnalysisUsage(AnalysisUsage &AU) const override;

	virtual StringRef getPassName() const override {
		return "Hello llvm";
	}

private:
	// Helper functions:
	// - Fetch or create a printf declaration
	// - Fetch or create a HelloWorld function which calls printf
	// - Fetch or create a main function which calls HelloWorld

	Function &GetPrintf(Module &Mod);
	Function &GetHelloWorld(Module &Mod, bool &Built);
	void AddMain(Module &Mod, Function &HelloFun, bool &Modified);

};

char HelloLLVM::ID = 0;

bool HelloLLVM::runOnModule(Module &Mod) {
	return true;
}

void HelloLLVM::getAnalysisUsage(AnalysisUsage &AU) const {
	// This pass does no require any analysis.
	// This pass potentially invalidates all analysis.
	// The default behaviour is to invalidate all, so no code is required here.
	return;
}

Function &HelloLLVM::GetHelloWorld(Module &Mod, bool &Modified) {
	Function *Fun = Mod.getFunction("hello_world");

	// Function already available in the Module.
	if(Fun && !Fun->isDeclaration()) {
		Modified = false;
		return *Fun;
	}

	// A context is used to track the compiler status -- it is required passing it
	// to function creating new LLVM objects in order to be inserted into compiler
	// state.
	LLVMContext &Ctx = Mod.getContext();

	Constant *FunSym;

	// No code associated to the function, first insert a declaration.
	FunSym = Mod.getOrInsertFunction("hello_world",		// fn name
																		Type::getVoidTy(Ctx), // fn type
																		NULL 						// attribute list
																	);
	Fun = cast<Function>(FunSym);

	// Then build a basic block to hold instructions.
	BasicBlock *Entry = BasicBlock::Create(Ctx, "entry", Fun);

	GlobalVariable *Msg;
	Value *MsgPtr;

	// We want to generate 'printf("Hello, World!")'. First parameter is a
	// constant.
	StringRef MsgContent("Hello, World!\n");
	Constant *MsgInit = ConstantDataArray::getString(Ctx, MsgContent);

	// Put it into a global variable.
	Msg = new GlobalVariable(Mod,
														MsgInit->getType(),
														true,
														GlobalVariable::InternalLinkage,
														MsgInit,
														"hello_world.msg");

	// Actually, 'printf' printf first parameter is a pointer -- we need to get
	// its address. Get an offsets vector first.
	SmallVector<Value *, 2> Offsets;
	Offsets.assign(2, ConstantInt::get(Type::getInt32Ty(Ctx), 0));

	// Then, use it to perform some pointer arithmetics.
	MsgPtr = GetElementPtrInst::Create(nullptr, Msg, Offsets, "", Entry);

	// Now we need the 'printf' function.
	Function &Printf = GetPrintf(Mod);

	// Call it, passing the string as the only argument.
	CallInst::Create(&Printf, MsgPtr, "", Entry);

	// A Basic block must always ends with a terminator instruction.
	// In this case a return.
	ReturnInst::Create(Ctx, Entry);

	Modified = true;
	return *Fun;
}

void HelloLLVM::AddMain(Module &Mod,
												Function &HelloFun,
												bool &Modified) {
	Function *Fun = Mod.getFunction("main");

	// Function already available.
	if(Fun && !Fun->isDeclaration()) {
		Modified = false;
		return;
	}

	LLVMContext &Ctx = Mod.getContext();
	Type *VoidTy = Type::getVoidTy(Ctx);

	Constant *FunSym;

	// Main not inside the compilation unit -- add it.
	FunSym = Mod.getOrInsertFunction("main", VoidTy, NULL);
	Fun = cast<Function>(FunSym);

	// Make an entry block.
	BasicBlock *Entry = BasicBlock::Create(Ctx, "entry", Fun);

	// Call the greeter and return.
	CallInst::Create(&HelloFun, ArrayRef<Value *>(), "", Entry);
	ReturnInst::Create(Ctx, Entry);

	Modified = true;
}

// The signature of 'printf' is architecture specific -- use this member
// function to define signatures with respect to the architecture.
Function &HelloLLVM::GetPrintf(Module &Mod) {
	LLVMContext &Ctx = Mod.getContext();
	FunctionType *PrintfTy;

#if defined(__i386__) || defined(__x86_64__)
	PrintfTy = FunctionType::get(Type::getInt32Ty(Ctx),
																Type::getInt8PtrTy(Ctx),
																true);
#else
	#error "Architecture not supported"
#endif

	Constant *FunSym = Mod.getOrInsertFunction("printf", PrintfTy);

	return *cast<Function>(FunSym);
}

// Pass registration: advise the PassManager about our pass
static RegisterPass<HelloLLVM> X(
			"Hello", 						// command line switch
			"Insert an Hello World Function",	// Human readable pass description
			true,								// True: Can modify CFG      | False: Only reads CFG
			true 								// True: Transformation Pass | False: Analysis Pass
			);

} /* cot */
